##############################################################################
#
# Copyright (c) 2005-2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Query terms for zc.catalog's SetIndex

$Id$
"""
from zc.catalog.interfaces import ISetIndex

from hurry.query import query


class SetTerm(query.IndexTerm):

    def getIndex(self, context):
        index = super(SetTerm, self).getIndex(context)
        assert ISetIndex.providedBy(index)
        return index


class All(SetTerm):

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'any': None})

    def key(self, context=None):
        return ('all', self.catalog_name, self.index_name)


class AnyOf(SetTerm):

    def __init__(self, index_id, values):
        super(AnyOf, self).__init__(index_id)
        self.values = tuple(values)

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'any_of': self.values})

    def key(self, context=None):
        return ('any of', self.catalog_name, self.index_name, self.values)


class AllOf(SetTerm):

    def __init__(self, index_id, values):
        super(AllOf, self).__init__(index_id)
        self.values = tuple(values)

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'all_of': self.values})

    def key(self, context=None):
        return ('all of', self.catalog_name, self.index_name, self.values)


class SetBetween(SetTerm):

    def __init__(self, index_id,
                 minimum=None, maximum=None,
                 exclude_min=False, exclude_max=False):
        super(SetBetween, self).__init__(index_id)
        self.options = (minimum, maximum, exclude_min, exclude_max)

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'between': self.options})

    def key(self, context=None):
        return ('between', self.catalog_name, self.index_name, self.options)


class ExtentAny(SetTerm):
    """Any ids in the extent that are indexed by this index."""

    def __init__(self, index_id, extent):
        super(ExtentAny, self).__init__(index_id)
        self.extent = extent

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'any': self.extent})


class ExtentNone(SetTerm):
    """Any ids in the extent that are not indexed by this index."""

    def __init__(self, index_id, extent):
        super(ExtentNone, self).__init__(index_id)
        self.extent = extent

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'none': self.extent})



##############################################################################
#
# Copyright (c) 2005-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Query terms for zc.catalog's ValueIndex

$Id$
"""
from zc.catalog.interfaces import IValueIndex

from hurry.query import query


class ValueTerm(query.IndexTerm):

    def getIndex(self, context):
        index = super(ValueTerm, self).getIndex(context)
        assert IValueIndex.providedBy(index)
        return index


class Eq(ValueTerm):

    def __init__(self, index_id, value):
        assert value is not None
        super(Eq, self).__init__(index_id)
        self.value = value

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'any_of': (self.value,)})

    def key(self, context=None):
        return ('equal', self.catalog_name, self.index_name, self.value)


class NotEq(ValueTerm):

    def __init__(self, index_id, value):
        super(NotEq, self).__init__(index_id)
        self.value = value

    def apply(self, cache, context=None):
        index = self.getIndex(context)
        values = list(index.values())
        # the remove method produces a value error when the value to
        # be removed is not in the list in the first place.  Having a
        # try/except clause is more efficent than first searching the
        # list for the value to remove.
        try:
            values.remove(self.value)
        except ValueError:
            pass
        return index.apply({'any_of': values})

    def key(self, context=None):
        return ('not equal', self.catalog_name, self.index_name, self.value)


class All(ValueTerm):

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'any': None})

    def key(self, context=None):
        return ('all', self.catalog_name, self.index_name)


class Between(ValueTerm):

    def __init__(self, index_id, min_value=None, max_value=None,
                 exclude_min=False, exclude_max=False):
        super(Between, self).__init__(index_id)
        self.options = (min_value, max_value, exclude_min, exclude_max)

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'between': self.options})

    def key(self, context=None):
        return ('between', self.catalog_name, self.index_name, self.options)


class Ge(Between):

    def __init__(self, index_id, min_value):
        super(Ge, self).__init__(index_id, min_value=min_value)


class Le(Between):

    def __init__(self, index_id, max_value):
        super(Le, self).__init__(index_id, max_value=max_value)


class In(ValueTerm):

    def __init__(self, index_id, values):
        assert None not in values
        super(In, self).__init__(index_id)
        self.values = tuple(values)

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'any_of': self.values})

    def key(self, context=None):
        return ('in', self.catalog_name, self.index_name, self.values)


class ExtentAny(ValueTerm):
    """Any ids in the extent that are indexed by this index."""

    def __init__(self, index_id, extent):
        super(ExtentAny, self).__init__(index_id)
        self.extent = extent

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'any': self.extent})


class ExtentNone(ValueTerm):
    """Any ids in the extent that are not indexed by this index."""

    def __init__(self, index_id, extent):
        super(ExtentNone, self).__init__(index_id)
        self.extent = extent

    def apply(self, cache, context=None):
        return self.getIndex(context).apply({'none': self.extent})



import functools
import time
import unittest
import zope.component.testing
import zope.intid.interfaces

from testfixtures import LogCapture
from zope.catalog.catalog import Catalog
from zope.catalog.field import FieldIndex
from zope.catalog.interfaces import ICatalog
from zope.catalog.text import TextIndex
from zope.component import provideUtility, getUtility
from zope.container.contained import Contained
from zope.interface import Interface, Attribute, implementer

from hurry.query import query
from hurry.query.interfaces import IQuery


"""Bring `query` testcoverage to 100% without polluting the doctest"""


class IContent(Interface):
    f1 = Attribute('f1')
    f2 = Attribute('f2')
    f3 = Attribute('f3')
    f4 = Attribute('f4')
    t1 = Attribute('t1')
    t2 = Attribute('t2')


@functools.total_ordering
@implementer(IContent)
class Content(Contained):

    def __init__(self, id, f1='', f2='', f3='', f4='', t1='', t2=''):
        self.id = id
        self.f1 = f1
        self.f2 = f2
        self.f3 = f3
        self.f4 = f4
        self.t1 = t1
        self.t2 = t2

    def __lt__(self, other):
        return self.id < other.id

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return '<Content "{}">'.format(self.id)


@implementer(zope.intid.interfaces.IIntIds)
class DummyIntId(object):
    MARKER = '__dummy_int_id__'

    def __init__(self):
        self.counter = 0
        self.data = {}

    def register(self, obj):
        intid = getattr(obj, self.MARKER, None)
        if intid is None:
            setattr(obj, self.MARKER, self.counter)
            self.data[self.counter] = obj
            intid = self.counter
            self.counter += 1
        return intid

    def getId(self, obj):
        return getattr(obj, self.MARKER)

    def getObject(self, intid):
        return self.data[intid]

    def __iter__(self):
        return iter(self.data)


f1 = ('catalog1', 'f1')


class QueryTestBase(unittest.TestCase):

    tearDown = zope.component.testing.tearDown

    def setUp(self):
        """emulate the doctest fixtures"""
        self.intid = DummyIntId()
        provideUtility(self.intid, zope.intid.interfaces.IIntIds)
        self.catalog = Catalog()
        provideUtility(self.catalog, ICatalog, 'catalog1')
        self.catalog['f1'] = FieldIndex('f1', IContent)
        self.catalog['f2'] = FieldIndex('f2', IContent)
        self.catalog['f3'] = FieldIndex('f3', IContent)
        self.catalog['f4'] = FieldIndex('f4', IContent)
        self.catalog['t1'] = TextIndex('t1', IContent)
        self.catalog['t2'] = TextIndex('t2', IContent)
        provideUtility(query.Query(), IQuery)
        self.setup_content()

    def setup_content(self):
        content = [
            Content(1, 'a', 'b', 'd'),
            Content(2, 'a', 'c'),
            Content(3, 'X', 'c'),
            Content(4, 'a', 'b', 'e'),
            Content(5, 'X', 'b', 'e'),
            Content(6, 'Y', 'Z')]
        for entry in content:
            self.catalog.index_doc(self.intid.register(entry), entry)

    def searchResults(self, q, **kw):
        query = getUtility(IQuery)
        return query.searchResults(q, **kw)

    def displayQuery(self, q, **kw):
        r = self.searchResults(q, **kw)
        return [e.id for e in sorted(list(r))]

    def test_setup(self):
        """verify test fixtures by reproducing first doctest"""
        self.assertEqual(self.displayQuery(
            query.All(f1)),
            [1, 2, 3, 4, 5, 6])


class TimingTest(QueryTestBase):

    def test_init(self):
        timer = query.Timing('foo', 1)
        self.assertEqual(timer.key, 'foo')
        self.assertGreater(timer.start, 0)
        self.assertEqual(timer.start_order, 1)
        self.assertEqual(timer.end, None)
        self.assertEqual(timer.end_order, None)

    def test_done(self):
        timer = query.Timing()
        timer.done(2)
        self.assertGreater(timer.end, 0)
        self.assertEqual(timer.end_order, 2)

    def test_total(self):
        timer = query.Timing()
        time.sleep(.1)
        timer.done()
        self.assertGreater(timer.total, 0)

    def test_total_wo_end(self):
        timer = query.Timing()
        self.assertEqual(timer.total, None)


class TimingAwareCacheTest(QueryTestBase):

    def test_init(self):
        cache = query.TimingAwareCache({})
        self.assertEqual(cache.cache, {})
        self.assertEqual(cache.timing, {})
        self.assertEqual(cache.count, 0)
        self.assertEqual(cache.post, None)

    def test_start_post(self):
        cache = query.TimingAwareCache({})
        cache.start_post()
        self.assertGreater(cache.post.start, 0)

    def test_end_post(self):
        cache = query.TimingAwareCache({})
        with self.assertRaises(AttributeError):
            cache.end_post()

        cache.start_post()
        time.sleep(0.1)
        cache.end_post()
        self.assertGreater(cache.post.total, 0)

    def test_get_uncached(self):
        cache = query.TimingAwareCache({})
        self.assertFalse('foo' in cache.timing)
        self.assertEqual(cache.count, 0)
        value = cache.get('foo')
        self.assertEqual(value, None)
        self.assertTrue('foo' in cache.timing)
        self.assertGreater(cache.timing['foo'].start, 0)
        self.assertEqual(cache.timing['foo'].end, None)
        self.assertEqual(cache.count, 1)

    def test_get_cached(self):
        cache = query.TimingAwareCache({'foo': 'bar'})
        self.assertFalse('foo' in cache.timing)
        self.assertEqual(cache.count, 0)
        value = cache.get('foo')
        self.assertEqual(value, 'bar')
        self.assertFalse('foo' in cache.timing)
        self.assertEqual(cache.count, 0)

    def test_dunder_setitem(self):
        cache = query.TimingAwareCache({})
        self.assertFalse('foo' in cache.timing)
        self.assertEqual(cache.count, 0)
        cache['foo'] = 'bar'
        self.assertFalse('foo' in cache.timing)
        self.assertEqual(cache.count, 0)

    def test_dunder_setitem_timing(self):
        cache = query.TimingAwareCache({})
        self.assertFalse('foo' in cache.timing)
        self.assertEqual(cache.count, 0)
        cache.get('foo')
        self.assertEqual(cache.count, 1)
        cache['foo'] = 'bar'
        self.assertTrue('foo' in cache.timing)
        self.assertGreater(cache.timing['foo'].start, 0)
        self.assertGreater(cache.timing['foo'].end, 0)
        self.assertEqual(cache.count, 2)

    def test_report_empty(self):
        with LogCapture() as logged:
            cache = query.TimingAwareCache({})
            cache.report()
            records = logged.records

        self.assertEqual(records, [])

    def test_report_uncached(self):
        with LogCapture() as logged:
            cache = query.TimingAwareCache({})
            cache.get('foo')
            cache['foo'] = 'bar'
            cache.report()
            records = logged.records

        self.assertEqual(records[0].levelname, 'INFO')
        self.assertEqual(records[0].module, 'query')
        self.assertIn('Catalog query', records[0].msg)
        self.assertIn('s for terms', records[0].msg)
        self.assertIn('s to finish', records[0].msg)
        self.assertIn('s: foo', records[1].msg)

    def test_report_uncached_post(self):
        with LogCapture() as logged:
            cache = query.TimingAwareCache({})
            cache.start_post()
            cache.get('foo')
            cache['foo'] = 'bar'
            cache.end_post()
            cache.report()
            records = logged.records

        self.assertEqual(records[0].levelname, 'INFO')
        self.assertEqual(records[0].module, 'query')
        self.assertIn('Catalog query', records[0].msg)
        self.assertIn('s for terms', records[0].msg)
        self.assertIn('s to finish', records[0].msg)
        self.assertIn('s: foo', records[1].msg)

    def test_report_uncached_post_under_over(self):
        with LogCapture() as logged:
            cache = query.TimingAwareCache({})
            cache.start_post()
            cache.get('foo')
            cache['foo'] = 'bar'
            cache.end_post()
            cache.report(over=1)
            records = logged.records

        self.assertEqual(records, [])

    def test_report_uncached_mixedup_order(self):
        with LogCapture() as logged:
            cache = query.TimingAwareCache({})
            cache.get('foobar')
            cache['foobar'] = 'foobar'
            cache.get('foo')
            cache.get('bar')
            cache['bar'] = 'bar'
            cache['foo'] = 'foo'
            cache.get('baz')
            cache['baz'] = 'baz'
            cache.report()
            records = logged.records

        # verify dedent from 5 spaces to 1 space
        self.assertEqual(len(records[1].msg) - len(records[1].msg.lstrip(' ')),
                         5)
        self.assertEqual(len(records[3].msg) - len(records[3].msg.lstrip(' ')),
                         1)

    def test_report_uncached_no_end_post(self):
        with LogCapture() as logged:
            cache = query.TimingAwareCache({})
            cache.start_post()
            cache.get('foo')
            cache.end_post()
            cache.get('foobar')
            cache['foo'] = 'bar'
            cache.report()
            records = logged.records

        self.assertTrue(records[2].msg.endswith('  ?: foobar.'))


class QueryTest(QueryTestBase):

    def test_injected_caching(self):
        class MockCaching(object):
            _cache = dict()
            _get = 0
            _set = 0

            def get(self, key):
                self._get += 1
                return self._cache.get(key)

            def __setitem__(self, key, value):
                self._set += 1
                self._cache[key] = value

        caching = MockCaching()
        self.searchResults(query.And(query.All(f1)), caching=caching)
        self.assertEqual(caching._get, 2)
        self.assertEqual(caching._set, 2)

        self.searchResults(query.And(query.All(f1)), caching=caching)
        self.assertEqual(caching._get, 3)
        self.assertEqual(caching._set, 2)
        self.assertEqual(
            sorted(caching._cache.keys()),
            [('all', 'catalog1', 'f1'), ('and', ('all', 'catalog1', 'f1'))])
        self.assertEqual(
            [v.keys() for v in caching._cache.values()],
            [[0, 1, 2, 3, 4, 5], [0, 1, 2, 3, 4, 5]])

    def test_timing(self):
        with LogCapture() as logged:
            self.searchResults(
                query.And(query.All(f1)), timing=.00000001)
            records = logged.records

        self.assertEqual(len(records), 3)
        self.assertIn("('and', ('all', 'catalog1', 'f1')", records[1].msg)
        self.assertIn("('all', 'catalog1', 'f1')", records[2].msg)

    def test_timing_cutoff(self):
        with LogCapture() as logged:
            self.searchResults(query.And(query.All(f1)), timing=5)
            records = logged.records

        self.assertEqual(len(records), 0)

    def test_timing_noresult(self):
        with LogCapture() as logged:
            results = self.displayQuery(
                query.And(query.Eq(f1, 'foo')), timing=.00000001)
            records = logged.records

        self.assertEqual(results, [])
        self.assertEqual(len(records), 3)


class TermsTest(QueryTestBase):

    def test_Term_apply(self):
        term = query.Term()
        with self.assertRaises(NotImplementedError):
            term.apply(None)

    def test_Term_dunder_rand(self):
        class AndDisabledAll(query.All):

            def __and__(self, other):
                return NotImplemented

        left = AndDisabledAll(f1)
        right = query.All(f1)

        self.assertEqual(self.displayQuery(
            left & right),
            [1, 2, 3, 4, 5, 6])

    def test_Term_dunder_ror(self):
        class OrDisabledAll(query.All):

            def __or__(self, other):
                return NotImplemented

        left = OrDisabledAll(f1)
        right = query.All(f1)

        self.assertEqual(self.displayQuery(
            left | right),
            [1, 2, 3, 4, 5, 6])

    def test_And_one_result(self):
        self.assertEqual(self.displayQuery(
            query.And(query.All(f1))),
            [1, 2, 3, 4, 5, 6])

    def test_And_empty_intersection(self):
        self.assertEqual(self.displayQuery(
            query.And(query.Eq(f1, 'a'), query.Eq(f1, 'X'))),
            [])

    def test_And_weighted(self):
        # this only executes the code path without any clue what the
        # impact of 'weighted' should be, if any
        self.assertEqual(self.displayQuery(
            query.And(query.All(f1), query.All(f1), weighted=True)),
            [1, 2, 3, 4, 5, 6])

    def test_And_weighted_empty_intersection(self):
        self.assertEqual(self.displayQuery(
            query.And(query.Eq(f1, 'a'), query.Eq(f1, 'X'), weighted=True)),
            [])

    def test_Or_one_empty_result(self):
        self.assertEqual(self.displayQuery(
            query.Or(query.Eq(f1, 'foo'))),
            [])

    def test_Or_one_result(self):
        self.assertEqual(self.displayQuery(
            query.Or(query.All(f1))),
            [1, 2, 3, 4, 5, 6])

    def test_Difference_one_empty_result(self):
        self.assertEqual(self.displayQuery(
            query.Difference(query.Eq(f1, 'foo'))),
            [])

    def test_Difference_second_empty_result(self):
        self.assertEqual(self.displayQuery(
            query.Difference(query.All(f1), query.Eq(f1, 'foo'))),
            [1, 2, 3, 4, 5, 6])

    def test_Difference_empty_difference(self):
        self.assertEqual(self.displayQuery(
            query.Difference(query.All(f1), query.All(f1))),
            [])

    def test_In_one_result(self):
        self.assertEqual(self.displayQuery(
            query.In(f1, ['Y', 'Z'])),
            [6])



##############################################################################
#
# Copyright (c) 2005-2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Basic query implementation

This module contains an IQuery utility implementation, basic query term
implementations and concrete term implementations for zope.catalog indexes.

$Id$
"""
import itertools
import logging
import os
import time

from BTrees.IFBTree import IFSet
from BTrees.IFBTree import weightedIntersection
from BTrees.IFBTree import multiunion, difference, intersection
from zope.cachedescriptors.property import Lazy
from zope.catalog.field import IFieldIndex
from zope.catalog.interfaces import ICatalog
from zope.catalog.text import ITextIndex
from zope.component import getUtility, getSiteManager, IComponentLookup
from zope.interface import implementer
from zope.intid.interfaces import IIntIds
from zope.index.interfaces import IIndexSort
from zope.index.text.parsetree import ParseError
from zope.location.location import located, LocationProxy
from hurry.query import interfaces


logger = logging.getLogger('hurry.query')
HURRY_QUERY_TIMING = 0.0  # log queries taking longer than this, in seconds
if 'HURRY_QUERY_TIMING' in os.environ:
    try:
        HURRY_QUERY_TIMING = float(os.environ['HURRY_QUERY_TIMING'])
    except (ValueError, TypeError):
        pass


class Locator(object):

    def __init__(self, container, get):
        self.container = container
        self.get = get

    def __call__(self, oid):
        contained = self.get(oid)
        return located(
            LocationProxy(contained), self.container, contained.__name__)


@implementer(interfaces.IResults)
class Results(object):

    def __init__(self, context, all_results, selected_results,
                 wrapper=None, locate_to=None):
        self.context = context
        self.locate_to = locate_to
        self.wrapper = wrapper
        self.__all = all_results
        self.__selected = selected_results

    @Lazy
    def get(self):
        get = getUtility(IIntIds, '', self.context).getObject
        if self.wrapper is not None:
            get = (lambda get: lambda id: self.wrapper(get(id)))(get)
        if self.locate_to is not None:
            return Locator(self.locate_to, get)
        return get

    @property
    def total(self):
        return len(self.__all)

    @property
    def count(self):
        return len(self.__selected)

    def first(self):
        for uid in self.__selected:
            return self.get(uid)

    def __len__(self):
        return len(self.__selected)

    def __iter__(self):
        for uid in self.__selected:
            yield self.get(uid)


@implementer(interfaces.IResults)
class NoResults(object):

    count = 0
    total = 0

    def first(self):
        return None

    def __len__(self):
        return 0

    def __iter__(self):
        return iter([])


no_results = NoResults()


class Timing(object):

    def __init__(self, key=None, order=0):
        self.key = key
        self.start = time.clock()
        self.start_order = order
        self.end = None
        self.end_order = None

    def done(self, order=0):
        self.end = time.clock()
        self.end_order = order

    @property
    def total(self):
        if self.end is not None:
            return self.end - self.start
        return None


class TimingAwareCache(object):

    def __init__(self, cache):
        self.cache = cache
        self.timing = {}
        self.count = 0
        self.post = None

    def start_post(self):
        self.post = Timing()

    def end_post(self):
        self.post.done()

    def __setitem__(self, key, value):
        self.cache[key] = value
        timing = self.timing.get(key)
        if timing is not None:
            timing.done(self.count)
            self.count += 1

    def get(self, key):
        value = self.cache.get(key)
        if value is None:
            self.timing[key] = Timing(key, self.count)
            self.count += 1
        return value

    def report(self, over=0):
        all_timing = sorted(self.timing.values(), key=lambda t: t.start_order)
        if not len(all_timing):
            return  # pragma: no cover (peephole optimizer interferes)
        total_post = 0 if self.post is None else self.post.total
        total_terms = all_timing[0].total
        if (total_terms + total_post) < over:
            return
        indent = 0
        order = [all_timing[0].end_order]
        logger.info(
            'Catalog query took {:.4f}s for terms, {:.4f}s to finish.'.format(
                total_terms, total_post))
        for timing in all_timing:
            if order == [] or timing.start_order < order[-1]:
                indent += 4
                order.append(timing.end_order)
            total = timing.total and '{:.4f}s'.format(timing.total) or '?'
            logger.info(
                '{} {}: {}.'.format(
                    ' ' * indent, total, str(timing.key)))
            if timing.end_order and len(order) \
               and timing.end_order > order[-1]:
                indent -= 4
                order.pop()


@implementer(interfaces.IQuery)
class Query(object):

    def searchResults(
            self, query, context=None, sort_field=None, limit=None,
            reverse=False, start=0, caching=None, timing=HURRY_QUERY_TIMING,
            wrapper=None, locate_to=None):

        if context is None:
            context = getSiteManager()
        else:
            context = IComponentLookup(context)

        if caching in (True, False, None):
            cache = {}
        else:
            # A custom cache object was injected, use it.
            cache = caching

        timer = None
        if timing:
            timer = cache = TimingAwareCache(cache)
        all_results = query.cached_apply(cache, context)
        if not all_results:
            if timer is not None:
                timer.report(over=timing)
            return no_results

        if timer is not None:
            timer.start_post()

        is_iterator = False
        if sort_field is not None:
            # Like in zope.catalog's searchResults we require the given
            # index to sort on to provide IIndexSort. We bail out if
            # the index does not.
            if not IIndexSort.providedBy(sort_field):
                assert isinstance(sort_field, tuple) and len(sort_field) == 2
                catalog_name, index_name = sort_field
                catalog = getUtility(ICatalog, catalog_name, context)
                sort_field = catalog[index_name]
                if not IIndexSort.providedBy(sort_field):
                    raise ValueError(
                        'Index {} in catalog {} does not support '
                        'sorting.'.format(index_name, catalog_name))
            sort_limit = None
            if limit:
                sort_limit = limit
                if start:
                    sort_limit += start
            selected_results = sort_field.sort(
                all_results,
                limit=sort_limit,
                reverse=reverse)
            if start:
                selected_results = itertools.islice(
                    selected_results, start, None)
            is_iterator = True
        else:
            # There's no sort_field given. We still allow to reverse
            # and/or limit the resultset. This mimics zope.catalog's
            # searchResults semantics.
            selected_results = all_results
            if reverse:
                selected_results = reversed(selected_results)
                is_iterator = True
            if limit or start:
                slice_end = limit and start + limit or None
                selected_results = itertools.islice(
                    selected_results, start, slice_end)
                is_iterator = True

        if is_iterator:
            selected_results = list(selected_results)

        if timer is not None:
            timer.end_post()
            timer.report(over=timing)

        return Results(
            context, all_results, selected_results, wrapper, locate_to)


@implementer(interfaces.ITerm)
class Term(object):

    def key(self, context=None):
        raise NotImplementedError()

    def apply(self, cache, context=None):
        raise NotImplementedError()

    def cached_apply(self, cache, context=None):
        try:
            key = self.key(context)
        except NotImplementedError:
            return self.apply(cache, context)
        cached = cache.get(key)
        if cached is not None:
            return cached
        result = self.apply(cache, context)
        cache[key] = result
        return result

    def __and__(self, other):
        return And(self, other)

    def __rand__(self, other):
        return And(other, self)

    def __or__(self, other):
        return Or(self, other)

    def __ror__(self, other):
        return Or(other, self)

    def __invert__(self):
        return Not(self)


class And(Term):

    def __init__(self, *terms, **kwargs):
        self.terms = terms
        self.weighted = kwargs.get('weighted', False)

    def apply(self, cache, context=None):
        results = []
        for term in self.terms:
            result = term.cached_apply(cache, context)
            if not result:
                # Empty results
                return result
            results.append(result)

        if len(results) == 1:
            return results[0]

        # Sort results to have the smallest set first to optimize the
        # set operation.
        results.sort(key=lambda r: len(r))

        result = results.pop(0)
        for r in results:
            if self.weighted:
                _, result = weightedIntersection(result, r)
            else:
                result = intersection(result, r)
            if not result:
                # Empty results
                return result

        return result

    def key(self, context=None):
        return ('and',) + tuple(term.key(context) for term in self.terms)


class Or(Term):

    def __init__(self, *terms):
        self.terms = terms

    def apply(self, cache, context=None):
        results = []
        for term in self.terms:
            result = term.cached_apply(cache, context)
            # empty results
            if not result:
                continue
            results.append(result)

        if len(results) == 0:
            return IFSet()
        if len(results) == 1:
            return results[0]

        return multiunion(results)

    def key(self, context=None):
        return ('or',) + tuple(term.key(context) for term in self.terms)


class Difference(Term):

    def __init__(self, *terms):
        self.terms = terms

    def apply(self, cache, context=None):
        results = []

        for index, term in enumerate(self.terms):
            result = term.cached_apply(cache, context)
            # If we do not have any results for the first index, just
            # return an empty set and stop here.
            if not result:
                if not index:
                    return IFSet()
                continue  # pragma: no cover (peephole optimizer interferes)
            results.append(result)

        result = results.pop(0)
        for other in results:
            result = difference(result, other)
            if not result:
                # Empty results
                return result
        return result

    def key(self, context=None):
        return ('difference',) + tuple(
            term.key(context) for term in self.terms)


class Not(Term):
    # XXX This term will load all the intids of your application
    # resulting in major and heavy performance issues. It is advised
    # not to use it.

    def __init__(self, term):
        self.term = term

    def apply(self, cache, context=None):
        return difference(self._all(), self.term.cached_apply(cache, context))

    def _all(self):
        return IFSet(uid for uid in getUtility(IIntIds))

    def key(self, context=None):
        return ('not', self.term.key(context))


class Objects(Term):

    def __init__(self, objects):
        self.objects = objects
        self._ids = None

    def ids(self, context=None):
        if self._ids is None:
            get_uid = getUtility(IIntIds, '', context).getId
            self._ids = tuple(get_uid(o) for o in self.objects)
        return self._ids

    def apply(self, cache, context=None):
        return IFSet(self.ids(context))

    def key(self, context=None):
        return ('objects', self.ids(context))


class Ids(Term):

    def __init__(self, *ids):
        self.ids = ids

    def apply(self, cache, context=None):
        return IFSet(self.ids)

    def key(self, context=None):
        return ('ids', self.ids)


class IndexTerm(Term):

    def __init__(self, catalog_name__and__index_name):
        self.catalog_name = catalog_name__and__index_name[0]
        self.index_name = catalog_name__and__index_name[1]

    def getIndex(self, context):
        catalog = getUtility(ICatalog, self.catalog_name, context)
        index = catalog[self.index_name]
        return index


class Text(IndexTerm):

    def __init__(self, index_id, text):
        super(Text, self).__init__(index_id)
        self.text = text

    def getIndex(self, context):
        index = super(Text, self).getIndex(context)
        assert ITextIndex.providedBy(index)
        return index

    def apply(self, cache, context=None):
        index = self.getIndex(context)
        try:
            return index.apply(self.text)
        except ParseError:
            logger.error(
                'search text "{}" yielded a ParseError'.format(self.text))
            return IFSet()

    def key(self, context=None):
        return ('text', self.catalog_name, self.index_name, self.text)


class FieldTerm(IndexTerm):

    def getIndex(self, context):
        index = super(FieldTerm, self).getIndex(context)
        assert IFieldIndex.providedBy(index)
        return index


class Eq(FieldTerm):

    def __init__(self, index_id, value):
        assert value is not None
        super(Eq, self).__init__(index_id)
        self.value = value

    def apply(self, cache, context=None):
        return self.getIndex(context).apply((self.value, self.value))

    def key(self, context=None):
        return ('equal', self.catalog_name, self.index_name, self.value)


class NotEq(FieldTerm):

    def __init__(self, index_id, value):
        super(NotEq, self).__init__(index_id)
        self.value = value

    def apply(self, cache, context=None):
        index = self.getIndex(context)
        values = index.apply((None, None))
        matches = index.apply((self.value, self.value))
        return difference(values, matches)

    def key(self, context=None):
        return ('not equal', self.catalog_name, self.index_name, self.value)


class All(FieldTerm):

    def apply(self, cache, context=None):
        return self.getIndex(context).apply((None, None))

    def key(self, context=None):
        return ('all', self.catalog_name, self.index_name)


class Between(FieldTerm):

    def __init__(self, index_id,
                 minimum=None, maximum=None):
        super(Between, self).__init__(index_id)
        self.options = (minimum, maximum)

    def apply(self, cache, context=None):
        return self.getIndex(context).apply(self.options)

    def key(self, context=None):
        return ('between', self.catalog_name, self.index_name, self.options)


class Ge(Between):

    def __init__(self, index_id, min_value):
        super(Ge, self).__init__(index_id, min_value, None)


class Le(Between):

    def __init__(self, index_id, max_value):
        super(Le, self).__init__(index_id, None, max_value)


class In(FieldTerm):

    def __init__(self, index_id, values):
        assert None not in values
        super(In, self).__init__(index_id)
        self.values = tuple(values)

    def apply(self, cache, context=None):
        results = []
        index = self.getIndex(context)
        for value in self.values:
            r = index.apply((value, value))
            # empty results
            if not r:
                continue
            results.append(r)

        if len(results) == 0:
            return IFSet()
        if len(results) == 1:
            return results[0]

        return multiunion(results)

    def key(self, context=None):
        return ('in', self.catalog_name, self.index_name, self.values)



##############################################################################
#
# Copyright (c) 2005-2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Query interfaces
"""

from zope.interface import Interface, Attribute


class IQuery(Interface):

    def searchResults(
            query, context=None, sort_field=None, limit=None, reverse=False,
            start=0, caching=False):

        """Query indexes.

        The query argument is a query composed of terms. Optionally
        provide the `context` parameter for the component lookups.

        Optionally provide a `sort_field` tuple that determines the
        index used to sort the result set with. This index is required
        to provide IIndexSort.

        Optionally provide a `limit` parameter to limit the result set
        to the given size.

        Optionally provide a `reverse` parameter to reverse the order
        of the result set.

        Optionally provide a `start` parameter to ignore the first
        result in set to the given start position.

        Optionally provide a `caching` parameter to cache terms result
        across multiple search queries. Accepted parameter values are:

            `True` for `threading.local` based and transaction boundary-aware
            caching. The cache for multiple `searchResults()` calls is
            invalidated at the end of the transaction.

            `False` to cache only for individual `searchResults()` calls.

            Or a `dict`-like object where invalidation is handled by the
            caller of the multiple `searchResults()` call.
        """


class ITerm(Interface):

    def key(context=None):
        """Return a unique key for this term.
        """

    def apply(cache, context=None):
        """Search and return the results for this term as an IFSet or
        something compatible with it.
        """

    def cached_apply(cache, context=None):
        """Look up in the cache and return results or apply the term if needed.
        """


class IResults(Interface):

    total = Attribute(
        'Total number of results (without start/limit restrictions)')

    count = Attribute(
        'Number of results (with start/limit restrictions)')

    def first():
        """Return only the first result of the query or None.
        """

    def __len__():
        """Return the number of results (with start/limit restrictions), that
        is the same than count.
        """

    def __iter__():
        """Iterate over the matching objects.
        """



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup

$Id$
"""
import os
import sys
from setuptools import setup, find_packages

if sys.version_info.major > 2:
    import html


def read(*rnames):
    text = open(os.path.join(os.path.dirname(__file__), *rnames)).read()
    if sys.version_info.major == 2:
        text = unicode(text, 'utf-8').encode('ascii', 'xmlcharrefreplace')
    else:
        text = html.escape(text)
    return text

tests_require = [
    'testfixtures',
    'zope.container',
    ]

setup(
    name="hurry.query",
    version='3.1',
    author='Infrae',
    author_email='faassen@startifact.com',
    description="Higher level query system for the zope.catalog",
    long_description=(read('src', 'hurry', 'query', 'query.txt') +
                      '\n\n' +
                      read('CHANGES.txt')),
    license='ZPL 2.1',
    keywords="zope zope3 catalog index query",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Framework :: Zope3'],
    url='http://pypi.python.org/pypi/hurry.query',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['hurry'],
    package_data={
        '': ['*.txt', '*.zcml'],
    },
    zip_safe=False,
    install_requires=[
        'BTrees',
        'setuptools',
        'transaction',
        'zc.catalog',
        'zope.cachedescriptors',
        'zope.catalog',
        'zope.component',
        'zope.index',
        'zope.interface',
        'zope.intid',
        'zope.location',
    ],
    tests_require=tests_require,
    extras_require={'test': tests_require},
    )
